unit TExecute;
// Direct3D Immediate mode test by Massimo Bardetti - bigben@dircon.co.uk.

// This unit implement TExecute a delphi wrapper for DX3 Execute buffers.
// Important :
// Always add vertices first.

interface
{$A+}{$Z4}{$T+}

uses Classes, Forms, SysUtils,  OLE2, Windows , DDraw, D3D, D3dTypes, D3DCaps;


type
  TExecuteBuffer = class ( TObject )
  private
   m_pID3D : IDirect3dDevice;
   m_pIViewPort : IDirect3dViewport;
   m_Desc : D3DEXECUTEBUFFERDESC;
   m_pIEB : IDirect3DExecuteBuffer;
   m_dwOffset : Integer;
   m_dwInsOffset : Integer;
   EBParent : TExecuteBuffer;
   lstEBChilds : TList;
   iExecuteOnce : integer;
   EBTemp : Pointer;

  public
    Tag : Integer;

//////////////////////////////////////////////////////////////////
//  Create the object with the following parameters :
//  pID3D : A direct3D device object
//  pIViewport : A Direct3d Viewport object
//  size : size in bytes of the execute buffer
    constructor Create ( pID3D : IDirect3dDevice; pIViewPort : IDirect3dViewport; size : integer ; pEbRoot : TExecuteBuffer) ;
    destructor Destroy ; override ;
//////////////////////////////////////////////////////////////////

// LockBuffer : lock the buffer in memory and return the pointer to the memory
// where it reside
    function LockBuffer : pointer;

// AddByte : Add one byte to the buffer
    function AddByte(b : BYTE) : Boolean;
// AddStruct : Add a portion of memory yo the buffer
    function AddStruct(pStruct : pointer; iSize : integer) : boolean;
// AddIns : Add D3D Op code
    function AddIns(op : D3DOPCODE; size : BYTE; count : WORD) : boolean;
// AddState : Add D3D State transformations (matrices)
// t : Type of transofrmation (world,projection, view ..);
// h : handle to the matrix describing it
    function AddState(t : D3DTRANSFORMSTATETYPE; h : D3DMATRIXHANDLE) : boolean;
// AddAmbientLight : Add Ambient Light!! clr is a 32 bit value ($RRGGBB)
    function AddAmbLight(clr : D3DColor) : Boolean;
// Execute : Unlocks and execute the buffer
    function Execute : Boolean;
// AddVertices : Add a set of vertices
// If your buffer will conatin vertices this is always the first instruction to do
// before adding anything else
// pVert : is a pointer to an array of pointers of D3DVERTEX type (eg. TList.List)
// nvert : Number of vertices
    function AddVertices(pVert : PPointerList; nVert : integer) : boolean;
// AddLitVertices : Add a set of vertices that has been lit and transformed by your
// code. This is useful if you want to create your own lighting and transformation code.
    function AddLitVertices(pVert : PPointerList; nVert : integer) : boolean;
// AddProcess : Add a process instruction
// nVert : number of vertices to process
// iStart : first vertex to process
// dwFlags : Flags (D3DPROCESSVERTICES_ type)
    function AddProcess(nVert : Integer; iStart : Integer; dwFlags : DWORD) : boolean;
// AddLines : Add a set of lines connecting two vertices
    function AddLines(var pLines : D3DLINE; nLine : Integer) : boolean;
// AddLine : Add a single line
    function AddLine(v1 : Integer; v2 : Integer) : boolean;
// AddPoint : Add point representing vertex
    function AddPoint(v : integer) : boolean;
// AddPoints : Add a set of points
    function AddPoints(var pPointInfo : D3DPOINT) : boolean;
// AddTriangle : Add Triangle connection v1,v2,v3 vertices
    function AddTriangle(v1,v2,v3 : Integer) : Boolean;
// AddMaterial : Add material handle
    function AddMaterial(hMat : D3DMATERIALHANDLE) : boolean;
// GetExecuteData : Fills a D3DEXECUTEDATA passed by reference
    function GetExecuteData(var ped : D3DEXECUTEDATA) : boolean;
// AddRenderState : Add a rander state instruction
    function AddRenderState(state : D3DRENDERSTATETYPE; Value : DWORD) : boolean;
    function AddLightState(state : D3DLIGHTSTATETYPE; Value : DWORD) : boolean;
// AddMatMul : Add a Matrix Multiplication to rotate/translate transformation matrices
    function AddMatMul(hDest, hSrc1, hSrc2 : D3DMATRIXHANDLE) : boolean;
// AddTriangleOperation : Add the Triangle operation and the number of tringles following it
// This function must always be followed by AddTriangleData to add iNum traingles to the
// execute buffer.
    function AddTriangleOperation(inum : Integer) : Boolean;
    function AddTriangleData(ptri : D3DTRIANGLE) : Boolean;
// Add a child buffer
   function AddChild(isize : integer) : TExecuteBuffer;
    procedure RemoveChild(EBCHild : TexecuteBuffer);
    function GetChild(i : Integer) : TExecuteBuffer;
    function NumChilds : Integer;

    function UnLockBuffer : HResult;

  end ;


implementation


var
   DDResult : HRESULT;

constructor TExecuteBuffer.Create ( pID3D : IDirect3dDevice; pIViewPort : IDirect3dViewport; size : integer ; pEbRoot : TExecuteBuffer) ;
begin
EBParent := pEBRoot;
if EBParent = nil then iExecuteOnce := 1
else iExecuteOnce := 2;

lstEBChilds := TList.Create;

m_pID3D := pID3D;
m_pID3D.AddRef;
m_pIViewport := pIViewPort;
m_pIViewPort.AddRef;

ZeroMemory(@m_Desc, sizeof(m_Desc));
m_Desc.dwSize := sizeof(m_Desc);
m_Desc.dwFlags := D3DDEB_BUFSIZE;
m_Desc.dwBufferSize := size;

m_pIEB := nil;
DDResult := m_pID3D.CreateExecuteBuffer(m_Desc,m_pIEB, nil);
if DDResult <> DD_OK then
    raise Exception.Create (
        Format ( 'CreateExecuteBuffer failed: %x', [ DDResult ] ) ) ;

if m_pIEB = nil then
    raise Exception.Create (
        Format ( 'CreateExecuteBuffer failed: %x', [ DDResult ] ) ) ;


m_dwOffset := 0;

AddByte(BYTE (D3DOP_EXIT));

m_dwOffset := 0;
m_dwInsOffset := 0;

end;

destructor TExecuteBuffer.Destroy ;
begin
     inherited;

     if (m_pIViewPort <> nil) then
        m_pIViewPort.Release;

     if (m_pIEB <> nil) then
        m_pIEB.Release;

     if (m_pID3D <> nil) then
        m_pID3D.Release;

     if EBParent <> nil then
        EBParent.RemoveChild(nil);
end;

function TExecuteBuffer.AddChild(isize : integer) : TExecuteBuffer;
var
   EBTemp : TExecuteBuffer;
begin
     EBTemp := TExecuteBuffer.Create(m_pID3D, m_pIViewport, iSize, self);

     try
     lstEBChilds.Add(EBTemp);
     except
     Result := nil;
     exit;
     end;

     Result := EBTemp;
     exit;
end;


procedure TExecuteBuffer.RemoveChild(EBChild : TExecuteBuffer);
var
   i : Integer;
begin
     if EBChild = nil then
        begin
             i := 0;
             while i < lstEBChilds.count  do
                   begin
                        RemoveChild(lstEBCHilds.Items[0]);
                        i := i + 1;
                   end;
             exit;
        end;

     EbChild.RemoveChild(nil);
     lstEBChilds.Remove(EBChild);
     EBChild.free();
end;

function TExecuteBuffer.GetChild(i : Integer) : TExecuteBuffer;
begin
     Result := lstEBChilds.Items[i];
end;

function TExecuteBuffer.NumChilds : Integer;
begin
     Result := lstEbChilds.Count;
end;

function TExecuteBuffer.LockBuffer : pointer;
begin
     if (m_Desc.lpdata = nil) then
        begin
             DdResult := m_pIEB.Lock(m_Desc);
             if DDResult <> DD_OK then
                 raise Exception.Create (
                       Format ( 'LockBuffer failed: %x', [ DDResult ] ) ) ;
        end;

     result := m_Desc.lpData;

end;

function TExecuteBuffer.UnLockBuffer : HResult;
begin
     DDResult := DD_OK;
     if m_Desc.lpData <> nil then
        begin
             DDResult := m_pIEB.UnLock();
             if (DDResult <> DD_OK) then
                Raise Exception.Create('Cannot Unlock');
             m_Desc.lpData := nil;
        end;

     Result := DDResult;
end;

function TExecuteBuffer.AddByte(b : BYTE) : Boolean;
var
   pBuf : PChar;
begin
     if (m_Desc.lpData = nil) then
        LockBuffer;

     pBuf := m_Desc.lpData;
     pBuf := pBuf + m_dwOffset;;

     pBuf^ := char (b);
     Inc(pBuf);
     Inc(m_dwOffset);
     pBuf^ := char (D3DOP_EXIT);

     Result := True;
end;

function TExecuteBuffer.AddStruct(pStruct : pointer; iSize : integer) : boolean;
var
   pSrc : pChar;
begin

     pSrc := pChar (pStruct);
     while (iSize > 0) do
           begin
                if (AddByte(BYTE (pSrc^)) = False) then
                 begin
                   Result := False;
                   exit;
                 end;

                Inc(pSrc);
                Dec(iSize);
           end;

     result := TRUE;
end;

function TExecuteBuffer.AddIns(op : D3DOPCODE; size : BYTE; count : WORD) : boolean;
var
   ins : D3DINSTRUCTION ;
   j : word;
begin
     ins.bOpCode := BYTE (op);
     ins.bSize := size;
     j := count;
     ins.wCount := count;

     Result := AddStruct(@ins, sizeof(ins));
end;

function TExecuteBuffer.AddState(t : D3DTRANSFORMSTATETYPE; h : D3DMATRIXHANDLE) : boolean;
var
   s : D3DSTATE;
begin
     if (not AddIns(D3DOP_STATETRANSFORM, sizeof(D3DSTATE), 1)) then
        begin
             result := False;
             exit;
        end;

     s.dtstTransformStateType := t;
     s.dwArg[0] := h;

     result := AddStruct(@s, sizeof(D3DSTATE));
end;

function TExecuteBuffer.AddAmbLight(clr : D3DColor) : Boolean;
var
   s : D3DSTATE;
begin
     if (not AddIns(D3DOP_STATELIGHT, sizeof(D3DSTATE), 1)) then
        begin
             Result := FAlse;
             exit;
        end;

     s.dlstLightStateType := D3DLIGHTSTATE_AMBIENT;
     s.dwArg[0] := clr;

     result := AddStruct(@s, sizeof(s));
end;

function TExecuteBuffer.Execute : Boolean;
label lbl1;
var
   ed : D3DEXECUTEDATA;
   i : Integer;
begin
     if iExecuteOnce = 0 then goto Lbl1;

     if m_Desc.lpData <> nil then
        begin
             DDResult := m_pIEB.UnLock();
             if (DDResult <> DD_OK) then
                Raise Exception.Create('Cannot Unlock');
             m_Desc.lpData := nil;
        end;

 //    ZeroMemory(@ed, sizeof(ed));
     ed.dwSize := sizeof(ed);
     ed.dwInstructionOffset := m_dwInsOffset;
     ed.dwInstructionLength := m_dwOffset - m_dwInsOffset + 1;
     ed.dwVertexCount := m_dwInsOffset div sizeof(D3DVERTEX);

     DDResult := m_pIEB.SetExecuteData(ed);
     if DDResult <> DD_OK then
        raise Exception.Create (
             Format ( 'SetExecuteData failed: %x', [ DDResult ] ) ) ;

lbl1:
     if EBParent = nil then
        begin
             DdResult := m_pID3D.BeginScene;
             if DDResult <> DD_OK then
                raise Exception.Create (
                      Format ( 'beginScene failed: %x', [ DDResult ] ) ) ;
        end;

     if iExecuteOnce <> 0 then
        begin
             DDResult := m_pID3D.Execute(m_pIEB, m_pIViewport, D3DEXECUTE_CLIPPED);
             if DDResult <> DD_OK then
             raise Exception.Create (
                   Format ( 'Execute failed: %x', [ DDResult ] ) ) ;
        end;

     if iExecuteOnce = 1 then
          iExecuteOnce := 0;

      i := 0;
     while i < lstEBChilds.Count do
           begin
                GetChild(i).Execute;
                i:=i+1;
           end;

     if EBParent = nil then
        begin
            DDResult := m_pID3D.EndScene;
             if DDResult <> DD_OK then
             raise Exception.Create (
                   Format ( 'EndScene failed: %x', [ DDResult ] ) ) ;
        end;


end;

function TExecuteBuffer.AddVertices(pVert : PPointerList; nVert : integer) : boolean;
var
   i : integer;
begin
     for i:=0 to nvert-1 do
         begin

              if (not AddStruct(pVert[i], sizeof(D3DVERTEX))) then
                 begin
                      Result := False;
                      exit;
                 end;
         end;

     m_dwInsOffset := m_dwOffset;
     result := True;
end;

function TExecuteBuffer.AddLitVertices(pVert : PPointerList; nVert : integer) : boolean;
var
   i : Integer;
begin
     for i:=0 to nvert-1 do
         begin

              if (not AddStruct(pVert[i], sizeof(D3DLVERTEX))) then
                 begin
                      Result := False;
                      exit;
                 end;
         end;

     m_dwInsOffset := m_dwOffset;
     result := True;
end;

function TExecuteBuffer.AddProcess(nVert : Integer; iStart : Integer; dwFlags : DWORD) : boolean;
var
   d : D3DPROCESSVERTICES;
begin
     if (not AddIns(D3DOP_PROCESSVERTICES, sizeof(D3DPROCESSVERTICES), 1)) then
        begin
             result := False;
             exit;
        end;


     ZeroMemory(@d,sizeof(d));
     d.dwFlags := dwFlags;
     d.wStart := iStart;
     d.wDest := d.wStart;
     d.dwCount := nVert;

     Result := AddStruct(@d, sizeof(d));
end;

function TExecuteBuffer.AddLines(var pLines : D3DLINE; nLine : Integer) : boolean;
begin
     if (not AddIns(D3DOP_LINE, sizeof(D3DLINE), nLine)) then
        begin
             Result := False;
             exit;
        end;

     Result := AddStruct(@pLines, nLine * sizeof(D3DLINE));

end;

function TExecuteBuffer.AddLine(v1 : Integer; v2 : Integer) : boolean;
var
   l : D3DLINE;
begin
     l.v1 := v1;
     l.v2 := v2;

     result := AddLines(l, 1);
end;

function TExecuteBuffer.AddPoint(v : integer) : boolean;
var
   p : D3DPOINT;
begin
     p.wCount := 1;
     p.wFirst := v;

     Result := AddPoints(p);
end;

function TExecuteBuffer.AddPoints(var pPointInfo : D3DPOINT) : boolean;
begin
     if (not AddIns(D3DOP_POINT, sizeof(D3DPOINT), 1)) then
      begin
        Result := False;
        exit;
      end;

     Result := AddStruct(@pPointInfo, sizeof(D3DPOINT));
end;

function  TExecuteBuffer.AddTriangle(v1,v2,v3 : Integer) : Boolean;
var
   t : D3DTRIANGLE;
begin
     if (not AddIns(D3DOP_TRIANGLE, sizeof(D3DTRIANGLE), 1)) then
        begin
             Result := False;
             exit;
        end;

     t.v1 := v1;
     t.v2 := v2;
     t.v3 := v3;
     t.wFlags := 0;

     Result := AddStruct(@t, sizeof(t));
end;

function  TExecuteBuffer.AddTriangleOperation(inum : Integer) : Boolean;
begin
     if (not AddIns(D3DOP_TRIANGLE, sizeof(D3DTRIANGLE), iNum)) then
        begin
             Result := False;
             exit;
        end;
end;

function  TExecuteBuffer.AddTriangleData(ptri : D3DTRIANGLE) : Boolean;
var
   t : D3DTRIANGLE;
begin
     t.v1 := ptri.v1;
     t.v2 := ptri.v2;
     t.v3 := ptri.v3;
     t.wFlags := ptri.wFlags;

     Result := AddStruct(@t, sizeof(t));
end;


function TExecuteBuffer.AddMaterial(hMat : D3DMATERIALHANDLE) : boolean;
var
   s : D3DSTATE;
begin
     if (not AddIns(D3DOP_STATELIGHT, sizeof(D3DSTATE), 1)) then
        begin
             Result := False;
             exit;
        end;

     s.dlstLightStateType := D3DLIGHTSTATE_MATERIAL;
     s.dwArg[0] := hMat;

     result := AddStruct(@s, sizeof(s));
end;

function TExecuteBuffer.GetExecuteData(var ped : D3DEXECUTEDATA) : boolean;
begin
     DDResult := m_pIEB.GetExecuteData(ped);
     if (DDResult <> DD_OK) then
        result := False
     else
         Result := True;

end;

function TExecuteBuffer.AddRenderState(state : D3DRENDERSTATETYPE; Value : DWORD) : boolean;
var
   s : D3DSTATE;
begin
     if (not AddIns(D3DOP_STATERENDER, sizeof(D3DSTATE), 1)) then
        begin
             result := false;
             exit;
        end;

     s.drstRenderStateType := state;
     s.dwArg[0] := Value;

     result := AddStruct(@s, sizeof(D3DSTATE));
end;

function TExecuteBuffer.AddLightState(state : D3DLIGHTSTATETYPE; Value : DWORD) : boolean;
var
   s : D3DSTATE;
begin
     if (not AddIns(D3DOP_STATELIGHT, sizeof(D3DSTATE), 1)) then
        begin
             result := false;
             exit;
        end;

     s.dlstLightStateType := state;
     s.dvArg[0] := Value;

     result := AddStruct(@s, sizeof(D3DSTATE));
end;


function TExecuteBuffer.AddMatMul(hDest, hSrc1, hSrc2 : D3DMATRIXHANDLE) : boolean;
var
   mm : D3DMATRIXMULTIPLY;
begin
     if (not AddIns(D3DOP_MATRIXMULTIPLY, sizeof(D3DMATRIXMULTIPLY), 1)) then
        begin
             result := false;
             exit;
        end;

     mm.hDestMatrix := hDest;
     mm.hSrcMatrix1 := hSrc1;
     mm.hSrcMatrix2 := hSrc2;

     Result := AddStruct(@mm, sizeof(mm));
end;


end.
